#-----------------------------------------------------------------------
# File    : coco4r.R
# Contents: COntinuous time ClOsed Neuron Assembly Detection for R
# Author  : Christian Borgelt
# History : 2015.08.05 file created (only function coconad)
#           2015.08.07 functions patspec() and estpsp() added
#           2015.08.12 function patred() added, library prefix added
#           2015.08.13 function psp2bdr() added
#           2016.03.23 perfect extension flag replaced by mode
#           2016.03.30 more target identifiers defined
#           2016.04.06 patterns changed to item set/support pairs
#           2016.10.05 list of target pattern types extended
#-----------------------------------------------------------------------

# target type identifiers
coco4r.targets <- c("si", "s", "set", "sets",
                    "i", "itemset", "itemsets",
                    "ai", "a", "all", "allset", "allsets",
                    "fi", "f", "frq", "freq", "frequent",
                    "frqset", "frqsets", "freqset", "freqsets",
                    "ci", "c", "cls", "clsd", "closed",
                    "mi", "m", "max", "maxi", "maximal",
                    "fp", "fo", "ap", "ao", "sp", "so",
                    "perm", "perms", "permutation", "permutations",
                    "allperm", "allperms",
                    "frqperm", "frqperms", "freqperm", "freqperms",
                    "cp", "co",
                    "clsperm", "clsperms", "clsdperm", "clsdperms",
                    "closedperm", "closedperms",
                    "mp", "mo",
                    "maxperm", "maxperms", "maxiperm", "maxiperms",
                    "maximalperm", "maximalperms",
                    "fs", "fq", "as", "aq", "ss", "sq",
                    "seq", "seqs", "sequence", "sequences",
                    "allseq", "allseqs",
                    "frqseq", "frqseqs", "freqseq", "freqseqs",
                    "cs", "cq",
                    "clsseq", "clsseqs", "clsdseq", "clsdseqs",
                    "closedseq", "closedseqs",
                    "ms", "mq",
                    "maxseq", "maxseqs", "maxiseq", "maxiseqs",
                    "maximalseq", "maximalseqs")

# CoCoNAD algorithm identifiers
coco4r.algos   <- c("b", "basic", "base",
                    "f", "filt", "filter", "filtered", "refined",
                    "r", "repo", "cmrepo", "repository")

# surrogate data generation identifiers
coco4r.surrs   <- c("i", "ident", "identity",
                    "r", "random", "randomize",
                    "d", "dither",
                    "s", "shift",
                    "k", "kernel", "kernest",
                    "p", "perm", "permute",
                    "b", "block", "blocked", "blkperm", "blockperm")

# density function identifiers
coco4r.rands   <- c("u", "uniform",
                    "r", "rect", "rectangle", "rectangular",
                    "t", "triang", "triangle", "triangular",
                    "g", "Gauss", "gauss",
                    "n", "normal")

# pattern set reduction function identifiers
coco4r.reds    <- c("x", "none",
                    "c", "coins",  "coins0",
                    "C",           "coins1", "coins+1",
                    "i", "items",  "items2",
                    "s", "cover",  "cover0", "covered", "covered0",
                    "S",           "cover1",            "covered1",
                    "l", "leni",   "leni0",  "lenient", "lenient0",
                    "L",           "leni1",             "lenient1",
                    "t", "strict", "strict0",
                    "T",           "strict1")

#-----------------------------------------------------------------------

coco4r.coconad <- function (trains, target="c", width=0.003, supp=2,
                            zmin=2, zmax=-1, report="a",
                            algo="r", mode="", border=NULL)
{                               # --- wrapper for CoCoNAD algorithm
  # check the function arguments: (only for debugging)
  #stopifnot(is.list(trains)       && (length(trains) > 0))
  #stopifnot(all(sapply(trains,is.numeric)))
  #stopifnot(is.character(target)  && any(target[1] == coco4r.targets))
  #stopifnot(is.numeric(width)     && (width >= 0))
  #stopifnot(is.numeric(supp))
  #stopifnot(is.numeric(zmin)      &&  (zmin >  0))
  #stopifnot(is.numeric(zmax)      && ((zmax <= 0) || (zmax > zmin)))
  #stopifnot(is.character(report)
  #  && any(report[1] == c("a","|","=","-")))
  #stopifnot(is.character(algo)    && any(algo[1]   == coco4r.algos))
  #stopifnot(is.character(mode))
  #stopifnot(is.null(border)       || is.numeric(border))
  # call the C implementation:
  r = .Call("c4r_coconad", trains, target, width, supp, zmin, zmax,
                           report, algo, mode, border)
  if (is.null(r)) warning("an error occurred, maybe out of memory")
  return(r)                     # check for correct execution
} # coco4r.coconad()

#-----------------------------------------------------------------------

coco4r.genpsp <- function (trains, target="c", width=0.003, supp=2,
                           zmin=2, zmax=-1, report="|",
                           algo="r", mode="",
                           cnt=1000, beg=-Inf, end=+Inf,
                           surr="p", rand="u", sigma=0.005, delta=0.03,
                           seed=0, cpus=0)
{                               # --- wrapper for pattern spectrum
  # check the function arguments:
  #stopifnot(is.list(trains)       && (length(trains) > 0))
  #stopifnot(all(sapply(trains,is.numeric)))
  #stopifnot(is.character(target)  && any(target[1] == coco4r.targets))
  #stopifnot(is.numeric(width)     && (width >= 0))
  #stopifnot(is.numeric(supp))
  #stopifnot(is.numeric(zmin)      && is.numeric(zmax))
  #stopifnot(is.character(report)
  #  && any(report[1] == c("#","|","=","-")))
  #stopifnot(is.character(algo)    && any(algo[1]   == coco4r.algos))
  #stopifnot(is.character(mode))
  #stopifnot(is.numeric(cnt))
  #stopifnot(is.numeric(beg)       && is.numeric(end))
  #stopifnot(is.character(surr)    && any(surr[1]   == coco4r.surrs))
  #stopifnot(is.character(rand)    && any(rand[1]   == coco4r.rands))
  #stopifnot(is.numeric(sigma)     && (sigma >= 0))
  #stopifnot(is.numeric(delta)     && (delta >  0))
  #stopifnot(is.numeric(seed)      && is.numeric(cpus))
  # call the C implementation:
  r = .Call("c4r_genpsp", trains, target, width, supp, zmin, zmax,
                          report, algo, mode, cnt, beg, end,
                          surr, rand, sigma, delta, seed, cpus)
  if (is.null(r)) warning("an error occurred, maybe out of memory")
  return(r)                     # check for correct execution
} # coco4r.genpsp()

#-----------------------------------------------------------------------

coco4r.estpsp <- function (trains, target="s", width=0.003, supp=2,
                           zmin=2, zmax=-1, report="|", equiv=10000,
                           alpha=0.5, smpls=1000, seed=0)
{                               # --- wrapper for pat. spec. estimation
  # check the function arguments:
  #stopifnot(is.list(trains)       && (length(trains) > 0))
  #stopifnot(all(sapply(trains,is.numeric)))
  #stopifnot(is.character(target)  && any(target[1] == coco4r.targets))
  #stopifnot(is.numeric(width)     && (width >= 0))
  #stopifnot(is.numeric(supp))
  #stopifnot(is.numeric(zmin)      && is.numeric(zmax))
  #stopifnot(is.character(report)
  #  && any(report[1] == c("#","|","=","-")))
  #stopifnot(is.numeric(equiv))
  #stopifnot(is.numeric(alpha)      && (alpha > 0))
  #stopifnot(is.numeric(smpls)      && (smpls > 0))
  #stopifnot(is.numeric(seed))
  # call the C implementation:
  r = .Call("c4r_estpsp", trains, target, width, supp, zmin, zmax,
                          report, equiv, alpha, smpls, seed)
  if (is.null(r)) warning("an error occurred, maybe out of memory")
  return(r)                     # check for correct execution
} # coco4r.estpsp()

#-----------------------------------------------------------------------

coco4r.psp2bdr <- function (psp)
{                               # --- extract the decision border
  stopifnot(is.list(psp))
  if ((length(psp) != 3)        # if list of triplets
  || ((length(psp) == 3) && all(sapply(psp, length) == 3)
  &&  (!is.integer(psp[[1]]) || !is.integer(psp[[2]]))))
    psp <- list(as.integer(sapply(psp, "[[", 1)),
                as.integer(sapply(psp, "[[", 2)))
  stopifnot(length(psp[[1]]) == length(psp[[2]]))
  stopifnot(is.numeric(psp[[1]])  && all(psp[[1]] >= 0))
  stopifnot(is.numeric(psp[[2]])  && all(psp[[2]] >= 0))
  b <- tapply(psp[[2]], psp[[1]], max)
  z <- min(psp[[1]])            # extract the decision border
  if (z > 0) {                  # if there are some sizes missing
    b <- c(rep(Inf, z), b+1)    # complete for sizes 0, 1 etc.
    names(b)[1:z] <- 0:(z-1)    # set names of added columns
  }
  for (i in (length(b)-1):1)    # ensure a monotone border
    if (b[i+1] > b[i]) b[i] = b[i+1]
  return(b)                     # return the created border
} # coco4r.psp2bdr()

#-----------------------------------------------------------------------

coco4r.patred <- function (pats, method="S", border=NULL, addis=TRUE)
{                               # --- wrapper for pat. set reduction
  # check the function arguments:
  stopifnot(is.list(pats))      # check the pattern types
  # stopifnot(all(sapply(pats, function(x) is.integer(x[[1]])))
  #        && all(sapply(pats, function(x) is.numeric(x[[2]]))))
  #stopifnot(is.character(method)  && any(method[1] == coco4r.reds))
  #stopifnot(is.null(border)       || is.numeric(border))
  #stopifnot(is.logical(addis))
  # call the C implementation:
  r = .Call("c4r_patred", pats, method, border, addis)
  if (is.null(r)) warning("an error occurred, maybe out of memory")
  return(r)                     # check for correct execution
} # coco4r.patred()
