/*----------------------------------------------------------------------
  File    : train.c
  Contents: train set management (realizations of linear point process)
  Author  : Christian Borgelt
  History : 2013.02.01 file created
            2013.02.02 file reading function tns_read() added
            2013.02.04 file writing sorted by points completed
            2013.02.06 linear minimum search for output of few items
            2013.02.19 functions trn_filterx(), trn_isect() etc. added
            2013.02.22 functions trn_filterx(), trn_isect() optimized
            2013.03.07 direction parameter added to sorting functions
            2013.03.19 function tns_recode() added (item reordering)
            2013.07.09 static constants emsgs and errmsgs merged
            2013.11.07 transaction reading capability made optional
            2013.11.12 item selection file changed to option -R#
            2013.12.02 surrogate generation functions added
            2013.12.03 main program extended by surrogate generation
            2013.12.21 points in tns_permute() taken from destination
            2014.05.09 function tns_permblk() added (blocked perms.)
            2014.07.11 NULL source check added to tna_filter/reduce()
            2014.10.24 changed from LGPL license to MIT license
            2016.10.12 special train/train array functions for sequences
            2016.10.13 function trn_seqisect() improved
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifdef TRN_MAIN
#ifndef TRN_SURR
#define TRN_SURR
#endif
#ifndef TRN_READ
#define TRN_READ
#endif
#ifndef TRN_WRITE
#define TRN_WRITE
#endif
#ifndef TA_READ
#define TA_READ
#endif
#endif
#include "train.h"
#ifdef TRN_SURR
#include "random.h"
#endif
#ifdef TRN_MAIN
#include "error.h"
#endif
#ifdef STORAGE
#include "storage.h"
#endif

#if defined _MSC_VER && !defined snprintf
#define snprintf    _snprintf
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "train"
#define DESCRIPTION "test train reading and writing functions"
#define VERSION     "version 1.13 (2014.10.24)        " \
                    "(c) 2013-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in tract.h */
#define E_STDIN      (-5)       /* double assignment of stdin */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* too few/many arguments */
#define E_RANDFN     (-9)       /* unknown random displacement func. */
#define E_SURRFN    (-10)       /* unknown surrogate generation meth. */
/* error codes -15 to -25 defined in tract.h */

#define BLKSIZE      256        /* block size for points/trains array */
#define HEAPMIN      15         /* minimum number of trains for heap */

#ifndef QUIET                   /* if not quiet version, */
#define MSG         fprintf     /* print messages */
#else                           /* if quiet version, */
#define MSG(...)    ((void)0)   /* suppress messages */
#endif

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

#ifndef CCHAR
#define CCHAR const char        /* abbreviation */
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- heap element --- */
  double     *p;                /* point value */
  const char *s;                /* name of item */
} HEAPE;                        /* (heap element) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef TRN_READ                 /* if train reading capability */
static CCHAR *errmsgs[] = {     /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_RANDFN   -9 */  "unkown random displacement function '%c'",
  /* E_SURRFN  -10 */  "unkown surrogate dat generation method '%c'",
  /*    -11 to -14 */  NULL, NULL, NULL, NULL,
  /* E_NOITEMS -15 */  "no (frequent) items found",
  /* E_ITEMEXP -16 */  "#item expected",
  /*           -17 */  NULL,
  /* E_DUPITEM -18 */  "#duplicate item '%s'",
  /*           -19 */  NULL,
  /* E_NUMEXP  -20 */  "#(floating point) number expected",
  /*           -21 */  NULL,
  /* E_FLDCNT  -22 */  "#too many fields/columns",
  /*           -23 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef TRN_MAIN
#ifndef QUIET
static CCHAR    *prgname;       /* program name for error messages */
#endif
static TABREAD  *tread  = NULL; /* table/train reader */
static TABWRITE *twrite = NULL; /* table/train writer */
static ITEMBASE *ibase  = NULL; /* item base */
static TRAINSET *trnset = NULL; /* train set */
static RNG      *rng    = NULL; /* random number generator */
#endif

#ifdef TRN_READ
static char msgbuf[2*TRD_MAXLEN+64];  /* buffer for error messages */
#endif

/*----------------------------------------------------------------------
  Train Functions
----------------------------------------------------------------------*/

TRAIN* trn_create (ITEM item, const double *pts, TID n)
{                               /* --- create a train */
  TRAIN *t;                     /* created train */

  assert(item >= 0);            /* check the function arguments */
  t = (TRAIN*)malloc(sizeof(TRAIN) +(size_t)n *sizeof(double));
  if (!t) return NULL;          /* allocate a new train and  */
  t->item = item;               /* store the corresponding item */
  t->size = n;                  /* and the size of the point array */
  t->pts[n] = TRN_END;          /* store a sentinel at the end */
  if (!pts)                     /* if no points, clear the length */
    t->pts[t->len = 0] = TRN_END;
  else                          /* if points are given, copy them */
    memcpy(t->pts, pts, (size_t)(t->len = n) *sizeof(double));
  return t;                     /* return the created train */
}  /* trn_create() */

/*--------------------------------------------------------------------*/

TRAIN* trn_resize (TRAIN *t, TID n)
{                               /* --- resize a train */
  assert(t);                    /* check the function arguments */
  if (t->size == n) return t;   /* check whether the size is right */
  t = realloc(t, sizeof(TRAIN) +(size_t)n *sizeof(double));
  if (!t) return NULL;          /* resize the memory block */
  t->size = n;                  /* store the new array size */
  if (n < t->len)               /* if to trim points, adapt length */
    t->pts[t->len = n] = TRN_END;
  return t;                     /* return the resized train */
}  /* trn_resize() */

/*--------------------------------------------------------------------*/

TRAIN* trn_clone (const TRAIN *t)
{                               /* --- clone a train */
  TRAIN *dst;                   /* created clone */

  assert(t);                    /* check the function arguments */
  dst = (TRAIN*)malloc(sizeof(TRAIN) +(size_t)t->len *sizeof(double));
  if (!dst) return NULL;        /* allocate a new train and */
  dst->item = t->item;          /* store the corresponding item */
  dst->size = dst->len = t->len;/* and the number of points */
  dst->pts[t->len] = TRN_END;   /* store a sentinel at the end */
  memcpy(dst->pts, t->pts, (size_t)t->len *sizeof(double));
  return dst;                   /* return the created train */
}  /* trn_clone() */

/*--------------------------------------------------------------------*/

void trn_copy (TRAIN *dst, const TRAIN *src)
{                               /* --- copy a train */
  assert(src                    /* check the function arguments */
  &&     dst && (dst->size >= src->len));
  dst->item = src->item;        /* copy the corresponding item */
  dst->len  = src->len;         /* and the number of points */
  dst->pts[src->len] = TRN_END; /* store a sentinel and copy points */
  memcpy(dst->pts, src->pts, (size_t)src->len *sizeof(double));
}  /* trn_copy() */

/*--------------------------------------------------------------------*/

int trn_add (TRAIN *t, double p)
{                               /* --- add a point to a train */
  if (t->len >= t->size) return -1;
  t->pts[  t->len] = p;         /* store the point */
  t->pts[++t->len] = TRN_END;   /* and a sentinel */
  return 0;                     /* return 'ok' */
}  /* trn_add() */

/*--------------------------------------------------------------------*/

int trn_cmp (const void *p1, const void *p2, void *data)
{                               /* --- compare trains by their item */
  ITEM n = (p1) ? ((const TRAIN*)p1)->item : ITEM_MAX;
  ITEM k = (p2) ? ((const TRAIN*)p2)->item : ITEM_MAX;
  if (n < k) return -1;         /* return the sign of the item */
  return (n > k) ? +1 : 0;      /* difference of the two trains */
}  /* trn_cmp() */

/*--------------------------------------------------------------------*/

int trn_cmpsz (const void *p1, const void *p2, void *data)
{                               /* --- compare trains by their size */
  ptrdiff_t n = (p1) ? (ptrdiff_t)((const TRAIN*)p1)->len : -1;
  ptrdiff_t k = (p2) ? (ptrdiff_t)((const TRAIN*)p2)->len : -1;
  if (n < k) return -1;         /* return the sign of the length */
  return (n > k) ? +1 : 0;      /* difference of the two trains */
}  /* trn_cmpsz() */

/*--------------------------------------------------------------------*/

TID trn_filter (TRAIN *trn, double beg, double end)
{                               /* --- filter train with an interval */
  double *s, *d;                /* to traverse the points */

  assert(trn && (beg <= end));  /* check the function arguments */
  for (s = d = trn->pts; *s < TRN_END; s++)
    if ((*s >= beg) && (*s <= end))
      *d++ = *s;                /* keep points in the interval */
  *d = TRN_END;                 /* store a sentinel point at the end */
  return trn->len = (TID)(d -trn->pts);
}  /* trn_filter() */           /* return the number of dest. points */

/*--------------------------------------------------------------------*/

TID trn_filterx (TRAIN *dst, TRAIN *src, IVAL *ivals)
{                               /* --- filter a train with intervals */
  double *s, *d;                /* to traverse the points */

  assert(dst && src && ivals);  /* check the function arguments */
  dst->item = src->item;        /* copy the associated item and */
  s = src->pts; d = dst->pts;   /* get source and destination points */
  if (ivals->beg < INFINITY) {  /* if there are intervals */
    for ( ; *s < TRN_END; s++){ /* train filtering loop */
      while (*s > ivals->end)   /* while point is after interval, */
        ivals++;                /* skip the current interval */
      if (ivals->beg >= INFINITY)
        break;                  /* if at end of intervals, abort */
      if (*s >= ivals->beg)     /* if point is in current interval, */
        *d++ = *s;              /* transfer it to the destination */
    }
  }
  *d = TRN_END;                 /* store a sentinel point at the end */
  return dst->len = (TID)(d-dst->pts);
}  /* trn_filterx() */          /* return the number of dest. points */

/*--------------------------------------------------------------------*/

TID trn_seqfiltx (TRAIN *dst, TRAIN *src, IVAL *ivals,
                  double width, int clomax)
{                               /* --- filter a train with intervals */
  double *s, *d;                /* to traverse the points */

  assert(dst && src && ivals);  /* check the function arguments */
  dst->item = src->item;        /* copy the associated item and */
  s = src->pts; d = dst->pts;   /* get source and destination points */
  if (ivals->beg < INFINITY) {  /* if there are intervals */
    for ( ; *s < TRN_END; s++){ /* train filtering loop */
      while (*s -ivals->beg > width)
        ivals++;                /* skip points after interval */
      if (ivals->beg >= INFINITY)
        break;                  /* if at end of intervals, abort */
      if ((*s >= ivals->beg)    /* if point is in current interval */
      || (clomax && ivals->end -*s <= width))
        *d++ = *s;              /* or may be added in before seq., */
    }                           /* transfer it to the destination */
  }
  *d = TRN_END;                 /* store a sentinel point at the end */
  return dst->len = (TID)(d-dst->pts);
}  /* trn_seqfiltx() */         /* return the number of dest. points */

/*--------------------------------------------------------------------*/

TID trn_isect (IVAL *dst, IVAL *src, TRAIN *train, double width)
{                               /* --- intersect intervals and train */
  TID    n = 0;                 /* number of used points of the train */
  double b, e;                  /* beginning and end of intersection */
  double *t;                    /* to traverse the train */
  IVAL   *d, *c;                /* to traverse the intervals */

  assert(dst                    /* check the function arguments */
  &&     src && train && (width >= 0));
  t = train->pts;               /* get the points of the train */
  if ((src->beg >= INFINITY) || (*t >= TRN_END)) {
    dst->beg = dst->end = INFINITY;
    return 0;                   /* if intervals or train are empty, */
  }                             /* store a sentinel and abort */
  c = d = dst;                  /* initialize the destination and */
  c->end = -INFINITY;           /* set the initial interval end */
  while (1) {                   /* interval list filtering loop */
    if (*t > src->end) {        /* if point is after current interval */
      if ((++src)->beg >= INFINITY)
        break; }                /* skip the processed interval */
    else {                      /* if point may be in curr. interval */
      if (*t >= src->beg) {     /* if point is in current interval */
        b = *t -width;          /* get beginning of intersection */
        if (src->beg > b) b = src->beg;
        e = *t +width;          /* get end       of intersection */
        if (src->end < e) e = src->end;
        if (b <= c->end) { c->end = e; }
        else { d->beg = b; d->end = e; c = d++; }
        n++;                    /* expand old or store new interval */
      }                         /* and count the used point */
      if (*++t >= TRN_END) break;
    }                           /* skip the processed point and */
  }                             /* check for the end of the train */
  d->beg = d->end = INFINITY;   /* store a sentinel interval */
  return n;                     /* return the number of used points */
}  /* trn_isect() */

/*--------------------------------------------------------------------*/

TID trn_seqisect (IVAL *dst, IVAL *src, TRAIN *train, double width)
{                               /* --- intersect intervals and train */
  double *t;                    /* to traverse the train */
  IVAL   *d;                    /* to traverse the output intervals */

  assert(dst                    /* check the function arguments */
  &&     src && train && (width >= 0));
  t = train->pts;               /* get the points of the train */
  if ((src->beg >= INFINITY) || (*t >= TRN_END)) {
    dst->beg = dst->end = INFINITY;
    return 0;                   /* if intervals or train are empty, */
  }                             /* store a sentinel and abort */
  d = dst;                      /* get the output interval array */
  if ((src->beg <= -INFINITY)   /* if there is only one interval that */
  &&  (src->end >= +INFINITY)){ /* spans the real line (first train) */
    while (*t < TRN_END) {      /* traverse the train */
      d->beg = d->end = *t++;   /* create one interval per point */
      d++;                      /* (minimum window covering sequence) */
    } }
  else {                        /* if this is not the first train */
    for ( ; src->beg < INFINITY; src++) {
      while (*t <= src->end)    /* skip all train times/points */
        t++;                    /* before the end of the window */
      if (*t -src->beg <= width) {
        d->beg = src->beg;      /* if time/point expands window, */
        d->end = *t; d++;       /* store an updated time interval */
      }                         /* in the output interval list */
    }
  }
  d->beg = d->end = INFINITY;   /* store a sentinel interval */
  return (TID)(d-dst);          /* return the number of intervals */
}  /* trn_seqisect() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void trn_show (TRAIN *t, ITEMBASE *base)
{                               /* --- show a train */
  double *p;                    /* to traverse the points */

  if (base) printf("%s/", ib_xname(base, t->item));
  printf("%"ITEM_FMT, t->item); /* print the associated item */
  for (p = t->pts; *p < TRN_END; p++)
    printf(" %g", *p);          /* traverse and print the points */
  printf(" [%"TID_FMT"]\n", t->len);
}  /* trn_show() */             /* finally print the number of points */

#endif
/*----------------------------------------------------------------------
  Surrogate Generation Functions
----------------------------------------------------------------------*/
#ifdef TRN_SURR

TRAIN* trn_ident (TRAIN *src, double beg, double end,
                  RNG *rng, TRAIN *dst)
{                               /* --- identity: original data */
  assert(src);                  /* check the function arguments */
  if (dst) trn_copy(dst, src);  /* copy the original train */
  else dst = trn_clone(src);    /* or clone it if there is no dest. */
  return dst;                   /* return the created surrogate */
}  /* trn_ident() */

/*--------------------------------------------------------------------*/

TRAIN* trn_random (TRAIN *src, double beg, double end,
                   RNG *rng, TRAIN *dst)
{                               /* --- randomize points/spikes */
  double *p;                    /* to traverse the points */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = trn_clone(src))) return NULL; }
  else if (dst != src)          /* if not identical to source, */
    trn_copy(dst, src);         /* copy the original train set */
  span = end -beg;              /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (p = dst->pts; *p != TRN_END; p++)
    *p = rng_dbl(rng) *span +beg;
  return dst;                   /* return the created surrogate */
}  /* trn_dither() */

/*--------------------------------------------------------------------*/

TRAIN* trn_dither (TRAIN *src, double beg, double end,
                   RNG *rng, TRAIN *dst)
{                               /* --- dither points/spikes */
  double *p, t;                 /* to traverse the points */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = trn_clone(src))) return NULL; }
  else if (dst != src)          /* if not identical to source, */
    trn_copy(dst, src);         /* copy the original train set */
  span = end -beg;              /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (p = dst->pts; *p != TRN_END; p++) {
    t  = fmod(rng_next(rng) +(*p -beg), span);
    *p = t +((t < 0) ? end : beg);
  }                             /* add a random value to each point */
  return dst;                   /* return the created surrogate */
}  /* trn_dither() */

/*--------------------------------------------------------------------*/

TRAIN* trn_shift (TRAIN *src, double beg, double end,
                  RNG *rng, TRAIN *dst)
{                               /* --- dither train/shift points */
  double *p, t;                 /* to traverse the points */
  double s;                     /* shift value for a train */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = trn_clone(src))) return NULL; }
  else if (dst != src)          /* if not identical to source, */
    trn_copy(dst, src);         /* copy the original train set */
  span = end -beg;              /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  s = rng_next(rng);            /* sample a random shift value */
  for (p = dst->pts; *p != TRN_END; p++) {
    t  = fmod(s +(*p -beg), span);
    *p = t +((t < 0) ? end : beg);
  }                             /* add a random shift to each point */
  return dst;                   /* return the created surrogate */
}  /* trn_shift() */

#endif
/*----------------------------------------------------------------------
  Train Array Functions
----------------------------------------------------------------------*/

ITEM tna_filter (TRAIN **dst, TRAIN **src, ITEM n,
                 IVAL *ivals, SUPP supp, void **mem)
{                               /* --- filter a train array */
  ITEM  i;                      /* loop variable for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    s = (SUPP)trn_filterx(t, src[i], ivals);
    if (s < supp) return -1;    /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_filter() */

/*--------------------------------------------------------------------*/

ITEM tna_seqfilt (TRAIN **dst, TRAIN **src, ITEM n,
                  IVAL *ivals, double width, SUPP supp, void **mem)
{                               /* --- filter a train array */
  ITEM  i;                      /* loop variable for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    s = (SUPP)trn_seqfiltx(t, src[i], ivals, width, 0);
    if (s < supp) return -1;    /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_seqfilt() */

/*--------------------------------------------------------------------*/

ITEM tna_filterx (TRAIN **dst, TRAIN **src, ITEM n,
                  IVAL *ivals, SUPP supp, ITEM *used, void **mem)
{                               /* --- filter a train array */
  ITEM  i, u;                   /* loop variable/index for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    if (used) {                 /* if first use index array is given */
      u = used[src[i]->item];   /* get first use index of source */
      if ((u >= 0) && (u < i)) { *d++ = dst[u]; continue; }
    }                           /* copy already filtered train */
    s = (SUPP)trn_filterx(t, src[i], ivals);
    if (s < supp) return -1;    /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_filterx() */

/*--------------------------------------------------------------------*/

ITEM tna_seqfiltx (TRAIN **dst, TRAIN **src, ITEM n, IVAL *ivals,
                   double width, SUPP supp, ITEM *used, void **mem)
{                               /* --- filter a train array */
  ITEM  i, u;                   /* loop variable/index for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    if (used) {                 /* if first use index array is given */
      u = used[trn_item(src[i])];
      if ((u >= 0) && (u < i)) { *d++ = dst[u]; continue; }
    }                           /* copy already filtered train */
    s = (SUPP)trn_seqfiltx(t, src[i], ivals, width, 0);
    if (s < supp) return -1;    /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_seqfiltx() */

/*--------------------------------------------------------------------*/

ITEM tna_reduce (TRAIN **dst, TRAIN **src, ITEM n,
                 IVAL *ivals, SUPP supp, void **mem)
{                               /* --- filter a train array */
  ITEM  i;                      /* loop variable for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    s = (SUPP)trn_filterx(t, src[i], ivals);
    if (s < supp) continue;     /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_reduce() */

/*--------------------------------------------------------------------*/

ITEM tna_seqred (TRAIN **dst, TRAIN **src, ITEM n, IVAL *ivals,
                 double width, SUPP supp, int clomax, void **mem)
{                               /* --- reduce a train array */
  ITEM  i;                      /* loop variable for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    s = (SUPP)trn_seqfiltx(t, src[i], ivals, width, clomax);
    if (s < supp) continue;     /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_seqred() */

/*--------------------------------------------------------------------*/

ITEM tna_seqredx (TRAIN **dst, TRAIN **src, ITEM n,
                  IVAL *ivals, double width, SUPP supp,
                  ITEM *used, TRAIN **trains, void **mem)
{                               /* --- reduce a train array */
  ITEM  i, u;                   /* loop variable/index for trains */
  SUPP  s;                      /* number of destination points */
  TRAIN **d, *t;                /* to traverse the trains */

  assert(dst                    /* check the function arguments */
  &&     src && ivals && (supp >= 0) && mem);
  t = *(TRAIN**)mem;            /* get the destination memory */
  for (d = dst, i = 0; i < n; i++) {
    if (!src[i]) continue;      /* traverse the source trains */
    if (used) {                 /* if first use index array is given */
      u = used[trn_item(src[i])];
      if (u >= 0) { *d++ = trains[u]; continue; }
    }                           /* copy already filtered train */
    s = (SUPP)trn_seqfiltx(t, src[i], ivals, width, 0);
    if (s < supp) continue;     /* filter and check against support */
    *d++ = t; t = (TRAIN*)(t->pts +(size_t)s +1);
  }                             /* store train and get next train */
  *(TRAIN**)mem = t;            /* store end of used memory */
  return (ITEM)(d-dst);         /* return the number of dest. trains */
}  /* tna_seqredx() */

/*----------------------------------------------------------------------
  Train Set Functions
----------------------------------------------------------------------*/

TRAINSET* tns_create (ITEMBASE *base)
{                               /* --- create a train set */
  TRAINSET *tns;                /* created train set */

  assert(base);                 /* check the function argument */
  tns = (TRAINSET*)malloc(sizeof(TRAINSET));
  if (!tns) return NULL;        /* create a train set */
  if (!base && !(base = ib_create(0, 0))) {
    free(tns); return NULL; }   /* create an item base if needed */
  tns->base   = base;           /* store the underlying item base */
  tns->min    =  INFINITY;      /* initialize the range of points */
  tns->max    = -INFINITY;
  tns->width  = 0.0;
  tns->maxseq = 0;
  tns->size   = tns->cnt = 0;
  tns->maxlen = 0;
  tns->extent = 0;
  tns->trains = NULL;           /* there are no trains yet */
  tns->pts    = NULL;           /* and no array of all points */
  tns->buf    = NULL;
  #ifdef TA_READ                /* if train reading capability */
  tns->err    = 0;              /* clear the error code and */
  tns->trd    = NULL;           /* the table/transaction reader */
  #endif
  return tns;                   /* return the created train set */
}  /* tns_create() */

/*--------------------------------------------------------------------*/

void tns_delete (TRAINSET *tns, int delib)
{                               /* --- delete a train set */
  assert(tns);                  /* check the function argument */
  if (tns->buf) free(tns->buf); /* delete a point heap */
  if (tns->pts) free(tns->pts); /* delete array of all points */
  if (tns->trains) {            /* if there are trains */
    while (--tns->cnt >= 0)     /* traverse the train array */
      if (tns->trains[tns->cnt]) free(tns->trains[tns->cnt]);
    free(tns->trains);          /* delete all trains */
  }                             /* and the train array */
  if (delib) ib_delete(tns->base);
  free(tns);                    /* delete the item base */
}  /* tns_delete() */           /* and the base structure */

/*--------------------------------------------------------------------*/

static TRAINSET* clone (TRAINSET *tns)
{                               /* --- clone train set structure */
  ITEM     i;                   /* loop variable */
  TRAINSET *dst;                /* created clone of the train set */

  dst = (TRAINSET*)malloc(sizeof(TRAINSET));
  if (!dst) return NULL;        /* create a train set */
  dst->trains = (TRAIN**)calloc((size_t)tns->cnt, sizeof(TRAIN*));
  if (!dst->trains) { free(dst); return NULL; }
  dst->base   = tns->base;      /* copy fields from the source */
  dst->size   = dst->cnt = tns->cnt;
  dst->maxlen = tns->maxlen;
  dst->extent = tns->extent;
  dst->pts    = NULL;           /* clear the array of all points */
  dst->buf    = NULL;
  for (i = 0; i < tns->cnt; i++) {
    dst->trains[i] = trn_create(tns->trains[i]->item, NULL,
                                tns->trains[i]->len);
    if (!dst->trains[i]) break; /* traverse and the trains and */
  }                             /* allocate destination memory */
  if (i < tns->cnt) {           /* if cloning the trains failed */
    for (i = 0; i < tns->cnt; i++)
      if (dst->trains[i]) trn_delete(dst->trains[i]);
    free(dst->trains); free(dst); return NULL;
  }                             /* clean up memory and abort */
  #ifdef TA_READ                /* if train reading capability */
  dst->err = 0;                 /* clear the error code and */
  dst->trd = NULL;              /* the table/transaction reader */
  #endif
  return dst;                   /* return the created train set */
}  /* clone() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_clone (TRAINSET *tns)
{                               /* --- clone a train set */
  TRAINSET *dst = clone(tns);   /* clone the memory structure */
  if (dst) tns_copy(dst, tns);  /* copy the trains into it */
  return dst;                   /* return the created clone */
}  /* tns_clone() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_copy (TRAINSET *dst, const TRAINSET *src)
{                               /* --- copy a train set */
  ITEM i;                       /* loop variable */

  assert(src                    /* check function arguments */
  &&     dst && (dst->size >= src->cnt));
  if (dst->pts) {               /* delete an existing point array */
    free(dst->pts); dst->pts = NULL; }
  for (i = 0; i < src->cnt; i++)/* traverse and copy the trains */
    trn_copy(dst->trains[i], src->trains[i]);
  dst->min = src->min;          /* copy the range of points */
  dst->max = src->max;          /* in the trains */
  return dst;                   /* return the destination */
}  /* tns_copy() */

/*--------------------------------------------------------------------*/

double tns_min (TRAINSET *tns, int recalc)
{                               /* --- get minimum point in trains */
  ITEM   i;                     /* loop variable for trains */
  double *p;                    /* to traverse the points */

  if ((tns->min <= -INFINITY) || (tns->min >= INFINITY) || recalc) {
    tns->min = INFINITY;        /* if to compute the minimum */
    for (i = 0; i < tns->cnt; i++) {
      for (p = tns->trains[i]->pts; *p != TRN_END; p++)
        if (*p < tns->min) tns->min = *p;
    }                           /* traverse trains and find */
  }                             /* minimum of the train points */
  return tns->min;              /* return the minimum point */
}  /* tns_min() */

/*--------------------------------------------------------------------*/

double tns_max (TRAINSET *tns, int recalc)
{                               /* --- get maximum point in trains */
  ITEM   i;                     /* loop variable for trains */
  double *p;                    /* to traverse the minimum points */

  if ((tns->max <= -INFINITY) || (tns->max >= INFINITY) || recalc) {
    tns->max = -INFINITY;       /* if to compute the maximum */
    for (i = 0; i < tns->cnt; i++) {
      for (p = tns->trains[i]->pts; *p != TRN_END; p++)
        if (*p > tns->max) tns->max = *p;
    }                           /* traverse trains and find */
  }                             /* maximum of the train points */
  return tns->max;              /* return the maximum point */
}  /* tns_max() */

/*--------------------------------------------------------------------*/

static void sift (const double *heap[], ITEM lft, ITEM rgt)
{                               /* --- let element sift down in heap */
  ITEM  i;                      /* index of first successor in heap */
  const double *t;              /* buffer for element to sift down */

  assert(heap && (lft >= 0) && (rgt > lft));
  t = heap[lft];                /* note the sift element */
  i = lft +lft +1;              /* compute index of first successor */
  do {                          /* sift loop */
    if ((i          < rgt)      /* if second successor exists and */
    &&  (*heap[i+1] < *heap[i]))/* it is smaller than the first, */
      i++;                      /* go to the second successor */
    if (*heap[i] >= *t)         /* if the successor is no less */
      break;                    /* than the sift element, abort */
    heap[lft] = heap[i];        /* let the successor ascend in heap */
    lft = i;                    /* note index of current element */
    i  += i+1;                  /* compute index of first successor */
  } while (i <= rgt);           /* while still within heap */
  heap[lft] = t;                /* store the sift element */
}  /* sift() */

/*--------------------------------------------------------------------*/

ITEM tns_maxseq (TRAINSET *tns, double width)
{                               /* --- find max. possible seq. length */
  ITEM   i, n;                  /* loop variable, number of trains */
  ITEM   curr;                  /* number of points in current window */
  double min;                   /* minimum point/event time */
  const double **heap;          /* heap for traversing the points */
  const double **tails;         /* tails of the trains (after window) */

  if ((width > 0) && (tns->width == width))
    return tns->maxseq;         /* if already known, return it */
  n     = tns->cnt;             /* get the number the trains and */
  width = fabs(width);          /* ensure non-negative window width */
  if (width <= 0)               /* if the window width is zero, */
    return tns->maxseq = n;     /* return the number of trains */
  heap = (const double**)malloc((size_t)(n+n) *sizeof(double));
  if (!heap) return -1;         /* create a train heap */
  tails = heap +n;              /* and a current tails array */
  for (i = 0; i < n; i++)       /* collect trains in heap and tail */
    heap[i] = tails[i] = tns->trains[i]->pts;
  for (i = n-- >> 1; --i >= 0; )
    sift(heap, i, n);           /* build the point heap */
  tns->maxseq = curr = 0;       /* initialize the number of points */
  while (*heap[0] < TRN_END) {  /* traverse the points of the trains */
    min = *heap[0];             /* get the minimum point/time */
    for (i = 0; i <= n; i++) {  /* traverse the trains */
      while ((*tails[i] < TRN_END)
      &&     (*tails[i] -min <= width)) {
        tails[i]++; curr++; }   /* advance the tail pointer(s) */
    }                           /* and count the covered points */
    if (curr > tns->maxseq)     /* if more points in this window, */
      tns->maxseq = curr;       /* update maximum number of points */
    heap[0]++; curr--;          /* skip the minimum point and */
    sift(heap, 0, n);           /* let the new point sift down */
  }                             /* (move window to next point) */
  assert(curr == 0);            /* last window must be empty */
  free((void*)heap);            /* delete the train heap */
  tns->width = width;           /* store the new window width */
  return tns->maxseq;           /* return maximum number of points */
}  /* tns_maxseq() */

/*--------------------------------------------------------------------*/

int tns_add (TRAINSET *tns, TRAIN *t)
{                               /* --- add a train to a train set */
  ITEM  i;                      /* loop variable, size buffer */
  TRAIN **p;                    /* to enlarge the train array */

  assert(tns && t               /* check the function arguments */
  &&    (t->item >= 0) && (t->item < ib_cnt(tns->base)));
  i = tns->size;                /* get the size of the train array */
  if (t->item >= i) {           /* if the train array is full */
    i += (i > BLKSIZE) ? i >> 1 : BLKSIZE;
    if (i <= t->item) i = t->item +1;
    p = (TRAIN**)realloc(tns->trains, (size_t)i *sizeof(TRAIN*));
    if (!p) return -1;          /* enlarge the train array */
    tns->trains = p; tns->size = i;
  }                             /* set enlarged array and its size */
  for (p = tns->trains +(i = tns->cnt); i < t->item; i++)
    *p++ = NULL;                /* clear intermediate trains */
  if ((t->item < tns->cnt) && tns->trains[t->item])
    return -2;                  /* check whether train already exists */
  tns->trains[t->item] = t;     /* store the new train */
  if (t->item >= tns->cnt)      /* adapt the number of trains */
    tns->cnt = t->item+1;       /* (depends on highest item) */
  if (t->len > tns->maxlen)     /* update the maximum length */
    tns->maxlen = t->len;       /* of a train and compute */
  tns->extent += (size_t)t->len;/* the total number of points */
  tns->min =  INFINITY;         /* invalidate the range of points */
  tns->max = -INFINITY;         /* (needs to be recomputed) */
  if (tns->pts) {               /* delete an existing point array */
    free(tns->pts); tns->pts = NULL; }
  return 0;                     /* return 'ok' */
}  /* tns_add() */

/*--------------------------------------------------------------------*/

void tns_sort (TRAINSET *tns, int dir)
{                               /* --- sort points in all trains */
  ITEM i;                       /* loop variable */

  assert(tns);                  /* check the function arguments */
  for (i = 0; i < tns->cnt; i++)/* sort the points in all trains */
    if (tns->trains[i]) trn_sort(tns->trains[i], dir);
}  /* tns_sort() */

/*--------------------------------------------------------------------*/

void tns_filter (TRAINSET *tns, double beg, double end)
{                               /* --- filter points in all trains */
  ITEM i;                       /* loop variable */

  assert(tns && (beg <= end));  /* check the function arguments */
  tns->extent = 0;              /* init. the total number of points */
  for (i = 0; i < tns->cnt; i++)/* filter the points in all trains */
    if (tns->trains[i])         /* and count the remaining points */
      tns->extent += (size_t)trn_filter(tns->trains[i], beg, end);
  tns->min =  INFINITY;         /* invalidate the range of points */
  tns->max = -INFINITY;         /* (needs to be recomputed) */
  if (tns->pts) {               /* delete an existing point array */
    free(tns->pts); tns->pts = NULL; }
}  /* tns_filter() */

/*--------------------------------------------------------------------*/

int tns_recode (TRAINSET *tns, int dir)
{                               /* --- recode items in trains */
  ITEM  i, k;                   /* loop variable, buffer */
  SUPP  s;                      /* number of points */
  TRAIN *t, **p;                /* to traverse the trains */
  ITEM  *map;                   /* identifier map for recoding */

  assert(tns);                  /* check the function arguments */
  for (i = 0; i < tns->cnt; i++) {
    t = tns->trains[i];         /* traverse the trains */
    s = (t) ? (SUPP)t->len : ((dir > 0) ? SUPP_MAX : -1);
    ib_setfrq(tns->base, i, s); /* set the item frequencies */
    ib_setxfq(tns->base, i, s); /* for sorting the trains */
  }
  p = (TRAIN**)malloc((size_t)tns->cnt *(sizeof(TRAIN)+sizeof(ITEM)));
  if (!p) return -1;            /* create a train buffer and */
  map = (ITEM*)(p+tns->cnt);    /* an item identifier map */
  ib_recode(tns->base, 0, -1, -1, dir, map);
  for (i = 0; i < tns->cnt; i++) {
    t = tns->trains[i];         /* traverse the trains again */
    p[k = map[i]] = t;          /* reorder the trains and */
    if (t) t->item = k;         /* map the item identifiers */
  }                             /* to their new values */
  memcpy(tns->trains, p, (size_t)tns->cnt *sizeof(TRAIN*));
  free(p);                      /* copy back the reordered trains */
  return 0;                     /* delete arrays and return 'ok' */
}  /* tns_recode() */

/*--------------------------------------------------------------------*/

double* tns_points (TRAINSET *tns)
{                               /* --- get an array with all points */
  ITEM   i;                     /* loop variable */
  double *s, *d;                /* to traverse the points */

  assert(tns);                  /* check the function arguments */
  if (tns->pts) return tns->pts;/* return an already existing array */
  tns->pts = d =(double*)malloc((size_t)(tns->extent+1)*sizeof(double));
  if (!d) return NULL;          /* create an array of all points */
  for (i = 0; i < tns->cnt; i++)/* traverse the trains */
    for (s = tns->trains[i]->pts; *s != TRN_END; )
      *d++ = *s++;              /* collect all points in the array */
  *d++ = TRN_END;               /* and store a sentinel at the end */
  return tns->pts;              /* return the created point array */
}  /* tns_points() */

/*--------------------------------------------------------------------*/
#ifdef TRN_READ

static TRAIN* trn_read (TRAINSET *tns, TABREAD *trd, int mode)
{                               /* --- read a train */
  int    d;                     /* delimiter type */
  ITEM   i;                     /* item identifier */
  char   *b, *e;                /* buffer for a field, end pointer */
  double p;                     /* buffer for a point value */
  TRAIN  *t, *x;                /* to access the trains, for realloc. */
  TID    z;                     /* size of point arrays in trains */
  char   buf[64];               /* buffer for item names */

  assert(trd                    /* check the function arguments */
  &&    ((mode == TRN_ITEMTRN) || (mode == TRN_PURETRN)));
  d = trd_read(trd);            /* read the next field */
  if (d <= TRD_ERR)     { tns->err = E_FREAD;   return NULL; }
  if (d <= TRD_EOF)     { tns->err = 1;         return NULL; }
  if (mode != TRN_ITEMTRN) sprintf(b = buf, "%"ITEM_FMT, tns->cnt+1);
  else                     b = trd_field(trd);
  if (!*b)              { tns->err = E_ITEMEXP; return NULL; }
  i = ib_item(tns->base, b);    /* get the item identifier */
  if ((i >= 0) && tns->trains[i]) { tns->err = E_DUPITEM; return NULL; }
  if ((i <  0)                  /* if the item is not known yet */
  &&  (ib_getapp(tns->base, -1) != APP_NONE)) {
    i = ib_add(tns->base, b);   /* add the item to the item base */
    if (i < 0)          { tns->err = E_NOMEM;   return NULL; }
  }
  if (i < 0) t = NULL;          /* create a new train (if needed) */
  else if (!(t = trn_create(i, NULL, BLKSIZE))) {
    tns->err = E_NOMEM; return NULL; }
  if (mode == TRN_ITEMTRN) {    /* if to parse a train with an item */
    if (d >= TRD_REC) return (t) ? trn_resize(t, 0) : NULL;
    d = trd_read(trd);          /* check for the end of the record */
  }                             /* and otherwise read the next field */
  z = 0;                        /* init. the point array size */
  while (1) {                   /* read the points of the train */
    b = trd_field(trd);         /* get the field that was read */
    if (!*b && (d >= TRD_REC)) break;
    p = strtod(b, &e);          /* convert field to a number */
    if (*e || (e == b)) {       /* and check for success */
      if (t) { trn_delete(t); } tns->err = E_NUMEXP; return NULL; }
    if (t && (t->len >= z)) {   /* if the point array is full */
      z += (z > BLKSIZE) ? z >> 1 : BLKSIZE;
      x = trn_resize(t, z);     /* enlarge the point array */
      if (!x) { if (t) {        /* and check for success */
        trn_delete(t); } tns->err = E_NOMEM; return NULL; }
      t = x;                    /* check for success and */
    }                           /* get the enlarged train */
    if (i >= 0) t->pts[t->len++] = p;
    if (d >= TRD_REC) break;    /* add the point to the train */
    d = trd_read(trd);          /* check for the end of the record, */
  }                             /* otherwise read the next field */
  if (!t) return NULL;          /* if no train read, abort */
  t->pts[t->len] = TRN_END;     /* store a sentinel at the end */
  return trn_resize(t, t->len); /* pack and return the read train */
}  /* trn_read() */

/*--------------------------------------------------------------------*/

int tns_read (TRAINSET *tns, TABREAD *trd, int mode)
{                               /* --- read trains from a file */
  int    d;                     /* delimiter type */
  ITEM   i, n;                  /* item identifier, counter */
  char   *b, *e;                /* buffer for a field, end pointer */
  double p = 0.0;               /* buffer for a point value */
  TRAIN  *t, *x, **a;           /* to access the trains, for realloc. */
  TID    z;                     /* size of point arrays in trains */

  assert(tns && trd);           /* check the function arguments */
  if (tns->pts) {               /* delete an existing point array */
    free(tns->pts); tns->pts = NULL; }
  tns->min =  INFINITY;         /* invalidate the range of points */
  tns->max = -INFINITY;         /* (needs to be recomputed) */
  tns->trd = trd;               /* note the table reader and */
  tns->err = 0;                 /* clear the error indicator */

  /* --- one train per record --- */
  if ((mode == TRN_ITEMTRN)     /* if one train per record, */
  ||  (mode == TRN_PURETRN)) {  /* maybe with a preceding item */
    while (1) {                 /* train read loop */
      t = trn_read(tns, trd, mode);
      if (tns->err > 0) break;  /* read the next train */
      if (tns->err < 0) { if (t) trn_delete(t); break; }
      if (t && (tns_add(tns, t) < 0)) { tns->err = E_NOMEM; break; }
    } }                         /* add the train to the train set */

  /* --- item/point pair per record --- */
  else {                        /* if pairs with item and point */
    while (1) {                 /* read list of item/point pairs */
      d = trd_read(trd);        /* read the next field */
      if (d <= TRD_ERR)     { tns->err = E_FREAD;   break; }
      if (d <= TRD_EOF)     { tns->err = 0;         break; }
      if (mode == TRN_PTITEM) { /* if points come first */
        b = trd_field(trd);     /* get the field that was read */
        p = strtod(b, &e);      /* and convert it to a number */
        if (*e || (e == b)) { tns->err = E_NUMEXP;  break; }
        d = trd_read(trd);      /* read the next field (item) */
        if (d <= TRD_ERR)   { tns->err = E_FREAD;   break; }
        if (d == TRD_FLD)   { tns->err = E_FLDCNT;  break; }
      }                         /* check read error/number of fields */
      b = trd_field(trd);       /* check whether an item was read */
      if (!*b)              { tns->err = E_ITEMEXP; break; }
      i = ib_item(tns->base,b); /* get the item identifier */
      if ((i < 0)               /* if the item is not known yet */
      &&  (ib_getapp(tns->base, -1) != APP_NONE)) {
        i = ib_add(tns->base,b);/* add the item to the item base */
        if (i < 0)          { tns->err = E_NOMEM;   break; }
      }
      if (mode == TRN_ITEMPT) { /* if items come first */
        if (d != TRD_FLD)   { tns->err = E_NUMEXP;  break; }
        d = trd_read(trd);      /* read the point value (a number) */
        if (d <= TRD_ERR)   { tns->err = E_FREAD;   break; }
        if (d == TRD_FLD)   { tns->err = E_FLDCNT;  break; }
        b = trd_field(trd);     /* get the field that was read */
        p = strtod(b, &e);      /* and convert it to a number */
        if (*e || (e == b)) { tns->err = E_NUMEXP;  break; }
      }                         /* check for a proper number */
      if (i < 0) continue;      /* skip new items to be ignored */
      if ((i < tns->cnt) && tns->trains[i]) {
        t = tns->trains[i];     /* if train already exists, get it */
        z = t->size;            /* get the current point array size */
        if (t->len >= z) {      /* if the point array is full */
          z += (z > BLKSIZE) ? z >> 1 : BLKSIZE;
          x = trn_resize(t, z); /* enlarge the train (point array) */
          if (!x)           { tns->err = E_NOMEM; break; }
          tns->trains[i] = t = x; t->size = z;
        } }                     /* store the enlarged train */
      else {                    /* if train does not yet exist */
        n = tns->size;          /* get the size of the train array */
        if (i >= n) {           /* if the train array is full */
          n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
          if (i > n) n = i+1;   /* compute new array size and */
          z = (TID)(n -tns->size);          /* added elements */
          a = realloc(tns->trains, (size_t)n *sizeof(TRAIN*));
          if (!a)           { tns->err = E_NOMEM; break; }
          memset(a +tns->size, 0, (size_t)z *sizeof(TRAIN*));
          tns->trains = a; tns->size = n;
        }                       /* enlarge the train array */
        tns->trains[i] = t = trn_create(i, NULL, BLKSIZE);
        if (!t)             { tns->err = E_NOMEM; break; }
        if (i >= tns->cnt) tns->cnt = i+1;
      }                         /* create and store a new train */
      t->pts[t->len++] = p;     /* store the new point in the train */
      t->pts[t->len]   = TRN_END;
    }                           /* store a sentinel after the points */
  }

  /* --- finalize the train set --- */
  tns->extent = 0;              /* clear the total size */
  tns->maxlen = 0;              /* and the maximum train length */
  for (i = 0; i < tns->cnt; i++) {
    t = tns->trains[i];         /* traverse the trains in the set, */
    if (!t) continue;           /* but skip missing trains */
    if (t->len > tns->maxlen)   /* update the maximum length */
      tns->maxlen = t->len;     /* of a train and compute the */
    tns->extent += (size_t)t->len;  /* total number of points */
    tns->trains[i] = trn_resize(t, t->len);
  }                             /* (try to) shrink the point array */
  return tns->err;              /* return the error status */
}  /* tns_read() */

/*--------------------------------------------------------------------*/

const char* tns_errmsg (TRAINSET *tns, char *buf, size_t size)
{                               /* --- get last (read) error message */
  int        i;                 /* index of error message */
  size_t     k = 0;             /* buffer for header length */
  const char *msg;              /* error message (format) */

  assert(tns                    /* check the function arguments */
  &&    (!buf || (size > 0)));  /* if none given, get internal buffer */
  if (!buf) { buf = msgbuf; size = sizeof(msgbuf); }
  i = (tns->err < 0) ? -tns->err : 0;
  assert(i < (int)(sizeof(errmsgs)/sizeof(char*)));
  msg = errmsgs[i];             /* get the error message (format) */
  assert(msg);                  /* check for a proper message */
  if (*msg == '#') { msg++;     /* if message needs a header */
    k = (size_t)snprintf(buf, size, "%s:%"SIZE_FMT"(%"SIZE_FMT"): ",
                         TRD_FPOS(tns->trd));
    if (k >= size) k = size-1;  /* print the input file name and */
  }                             /* the record and field number */
  snprintf(buf+k, size-k, msg, trd_field(tns->trd));
  return buf;                   /* format the error message */
}  /* tns_errmsg() */

#endif
/*--------------------------------------------------------------------*/
#ifdef TRN_WRITE

static void xsift (HEAPE *heap, ITEM lft, ITEM rgt)
{                               /* --- let element sift down in heap */
  ITEM   i;                     /* index of first successor in heap */
  HEAPE  t;                     /* buffer for element to sift down */
  double x;                     /* buffer for successor point */

  t = heap[lft];                /* note sift element */
  i = lft +lft +1;              /* compute index of first successor */
  do {                          /* sift loop */
    x = *heap[i].p;             /* get point of first sucecssor */
    if ((i            < rgt)    /* if second successor exists and */
    &&  (*heap[i+1].p < x))     /* it is smaller than the first, */
      x = *heap[++i].p;         /* go to the second successor */
    if (x >= *t.p)              /* if the successor is no less */
      break;                    /* than the sift element, abort */
    heap[lft] = heap[i];        /* let the successor ascend in heap */
    lft = i;                    /* note index of current element */
    i  += i+1;                  /* compute index of first successor */
  } while (i <= rgt);           /* while still within heap */
  heap[lft] = t;                /* store the sift element */
}  /* xsift() */

/*--------------------------------------------------------------------*/

static void pairout (TABWRITE *twr, int mode, const char *fmt, HEAPE *h)
{                               /* --- write item/point pair */
  assert(twr && fmt && h);      /* check the function arguments */
  if (mode == TRN_PTITEM) {     /* if to print the point first */
    twr_printf(twr, fmt, *h->p); twr_fldsep(twr);
    twr_puts(twr, h->s);         twr_recsep(twr); }
  else {                        /* if to print the item first */
    twr_puts(twr, h->s);         twr_fldsep(twr);
    twr_printf(twr, fmt, *h->p); twr_recsep(twr); }
}  /* pairout() */

/*--------------------------------------------------------------------*/

int tns_write (TRAINSET *tns, TABWRITE *twr, int mode, const char *fmt)
{                               /* --- write trains to a file */
  int    m;                     /* mode without flag TRN_SORTED */
  ITEM   i, k;                  /* loop variable, buffer */
  TRAIN  *t;                    /* to traverse the trains */
  double *p;                    /* to traverse the points */
  HEAPE  *heap;                 /* heap for sorted output */
  const char *s;                /* buffer for an item name */

  assert(tns && twr);           /* check the function arguments */
  m = mode & ~TRN_SORTED;       /* get the pure write mode */
  if (!(mode & TRN_SORTED)      /* if to split by items */
  ||  (tns->cnt <= 1)           /* or there is only one item */
  ||  (m == TRN_ITEMTRN)        /* or one train per record */
  ||  (m == TRN_PURETRN)) {     /* (with or w/o heading item) */
    for (i = 0; i < tns->cnt; i++) {
      t = tns->trains[i];       /* traverse the trains */
      s = ib_xname(tns->base, t->item);
      if ((m == TRN_ITEMTRN)    /* if to write one train per record */
      ||  (m == TRN_PURETRN)) { /* (with or w/o heading item) */
        if (m == TRN_ITEMTRN) { /* write the item if requested */
          twr_puts(twr, s); twr_fldsep(twr); }
        for (p = t->pts; *p < TRN_END; p++) {
          twr_printf(twr, fmt, *p);
          if (p[1] == TRN_END) twr_recsep(twr);
          else                 twr_fldsep(twr);
        } }                     /* write a list of points */
      else {                    /* if to write one pair per record */
        for (p = t->pts; *p < TRN_END; p++) {
          if (mode == TRN_PTITEM) { /* if to print the point first */
            twr_printf(twr, fmt, *p); twr_fldsep(twr);
            twr_puts(twr, s);         twr_recsep(twr); }
          else {                /* if to print the item first */
            twr_puts(twr, s);         twr_fldsep(twr);
            twr_printf(twr, fmt, *p); twr_recsep(twr); }
        }                       /* print pair of item and point */
      }
    } }
  else {                        /* if to write sorted by point */
    heap = (HEAPE*)malloc((size_t)tns->cnt *sizeof(HEAPE));
    if (!heap) return E_NOMEM;  /* create a heap */
    for (i = 0; i < tns->cnt; i++) {
      t = tns->trains[i];       /* traverse the trains */
      heap[i].s = ib_name(tns->base, t->item);
      heap[i].p = t->pts;       /* copy point lists to the heap */
    }                           /* together with the item names */
    if (tns->cnt > HEAPMIN) {   /* if there are many items */
      for (k = tns->cnt-1, i = tns->cnt >> 1; --i >= 0; )
        xsift(heap, i, k);      /* build the point heap */
      while (*heap[0].p < TRN_END) {
        pairout(twr, m, fmt, heap);
        heap[0].p++;            /* report top element of heap, */
        xsift(heap, 0, k);      /* skip the reported point and */
      } }                       /* let the new point sift down */
    else if (tns->cnt > 2) {    /* if there are few items */
      while (1) {               /* item/point pair write loop */
        for (i = k = 0; ++i < tns->cnt; )
          if (*heap[i].p < *heap[k].p)
             k = i;             /* find train with minimum point */
        if (*heap[k].p >= TRN_END)
          break;                /* if all trains empty, abort loop */
        pairout(twr, m, fmt, heap+k);
        heap[k].p++;            /* write the minimum item/point pair */
      } }                       /* and then skip the processed pair */
    else {                      /* if there are only two items */
      while ((*heap[0].p < TRN_END)  /* while there are points left */
      ||     (*heap[1].p < TRN_END)) {
        if (*heap[0].p < *heap[1].p)
             { pairout(twr, m, fmt, heap+0); heap[0].p++; }
        else { pairout(twr, m, fmt, heap+1); heap[1].p++; }
      }                         /* report the item with smaller point */
    }                           /* and then skip the processed pair */
    free(heap);                 /* delete the created heap */
  }
  return (twr_error(twr)) ? E_FWRITE : 0;
}  /* tns_write() */            /* return a write error indicator */

#endif
/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void tns_show (TRAINSET *tns)
{                               /* --- show a train set */
  ITEM   i;                     /* loop variable */
  TRAIN  *t;                    /* to traverse the trains */
  double *p;                    /* to traverse the points */
  const char *s;                /* buffer for an item name */

  assert(tns);                  /* check the function arguments */
  for (i = 0; i < tns->cnt; i++) {
    t = tns->trains[i];         /* traverse the trains */
    s = ib_xname(tns->base, t->item);
    for (p = t->pts; *p != TRN_END; p++)
      printf("%s %g\n", s, *p); /* print point, field separator, */
  }                             /* item, and record separator */
}  /* tns_show() */

#endif
/*----------------------------------------------------------------------
  Surrogate Generation Functions
----------------------------------------------------------------------*/
#ifdef TRN_SURR

TRAINSET* tns_ident (TRAINSET *src, double beg, double end,
                     double delta, RNG *rng, TRAINSET *dst)
{                               /* --- identity: original data */
  assert(src && (beg <= end));  /* check the function arguments */
  if (!dst) return tns_clone(src);
  return dst;                   /* return the created surrogate */
}  /* tns_ident() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_random (TRAINSET *src, double beg, double end,
                      double delta, RNG *rng, TRAINSET *dst)
{                               /* --- randomize points/spikes */
  ITEM   i;                     /* loop variable for items */
  double *p;                    /* to traverse the points */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = clone(src))) return NULL; }
  if (dst != src)               /* if not identical to source, */
    tns_copy(dst, src);         /* copy the original train set */
  span = end -beg;              /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (i = 0; i < dst->cnt; i++)/* traverse the trains */
    for (p = dst->trains[i]->pts; *p != TRN_END; p++)
      *p = rng_dbl(rng) *span +beg;
  dst->min =  INFINITY;         /* compute new points in span */
  dst->max = -INFINITY;         /* invalidate the range of points */
  return dst;                   /* return the created surrogate */
}  /* tns_random() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_dither (TRAINSET *src, double beg, double end,
                      double delta, RNG *rng, TRAINSET *dst)
{                               /* --- dither points/spikes */
  ITEM   i;                     /* loop variable for items */
  double *p, t;                 /* to traverse the points */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = clone(src))) return NULL; }
  if (dst != src)               /* if not identical to source, */
    tns_copy(dst, src);         /* copy the original train set */
  span = end -beg;              /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (i = 0; i < dst->cnt; i++) { /* traverse the trains */
    for (p = dst->trains[i]->pts; *p != TRN_END; p++) {
      t  = fmod(rng_next(rng) +(*p -beg), span);
      *p = t +((t < 0) ? end : beg);
    }                           /* add a random value to each point */
  }                             /* and wrap round if outside interval */
  dst->min =  INFINITY;         /* invalidate the range of points */
  dst->max = -INFINITY;         /* in the train set */
  return dst;                   /* return the created surrogate */
}  /* tns_dither() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_shift (TRAINSET *src, double beg, double end,
                     double delta, RNG *rng, TRAINSET *dst)
{                               /* --- dither train/shift points */
  ITEM   i;                     /* loop variable for items */
  double *p, t;                 /* to traverse the points */
  double s;                     /* shift value for a train */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = clone(src))) return NULL; }
  if (dst != src)               /* if not identical to source, */
    tns_copy(dst, src);         /* copy the original train set */
  span = end -beg;              /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (i = 0; i < dst->cnt; i++) {
    s = rng_next(rng);          /* traverse the trains */
    for (p = dst->trains[i]->pts; *p != TRN_END; p++) {
      t  = fmod(s +(*p -beg), span);
      *p = t +((t < 0) ? end : beg);
    }                           /* add a random shift to each point */
  }                             /* and wrap round if outside interval */
  dst->min =  INFINITY;         /* invalidate the range of points */
  dst->max = -INFINITY;         /* in the train set */
  return dst;                   /* return the created surrogate */
}  /* tns_shift() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_kernest (TRAINSET *src, double beg, double end,
                       double delta, RNG *rng, TRAINSET *dst)
{                               /* --- sample from kernel estimate */
  ITEM   i;                     /* loop variable for items/trains */
  SUPP   k;                     /* loop variable for points */
  size_t n, x;                  /* loop variables for points, index */
  double *p, t;                 /* to traverse the points */
  double *pts;                  /* array of all points to sample from */
  double span;                  /* length of the point span */

  assert(src && (beg <= end) && rng); /* check function arguments */
  pts = tns_points(src);        /* get an array with all points */
  if (!pts) return NULL;        /* as a point sampling source */
  if (!dst) { if (!(dst = clone(src))) return NULL; }
  n    = tns_extent(src);       /* get the total number of points */
  span = end -beg;              /* and compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (i = 0; i < dst->cnt; i++) {
    dst->trains[i]->len = src->trains[i]->len;
    p = dst->trains[i]->pts;    /* traverse the trains */
    for (k = 0; k < dst->trains[i]->len; k++) {
      x = (size_t)(rng_dbl(rng) *(double)n);
      if (x >= n) x = n-1;      /* get a random index in {0,...,n-1} */
      t = fmod(rng_next(rng) +(pts[x] -beg), span);
      *p++ = t +((t < 0) ? end : beg);
    }                           /* sample from the kernel function */
    *p = TRN_END;               /* and wrap round if outside interval */
  }                             /* finally store a sentinel */
  dst->min =  INFINITY;         /* invalidate the range of points */
  dst->max = -INFINITY;         /* in the train set */
  return dst;                   /* return the created surrogate */
}  /* tns_kernest() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_permute (TRAINSET *src, double beg, double end,
                       double delta, RNG *rng, TRAINSET *dst)
{                               /* --- dither permutation of points */
  ITEM   i;                     /* loop variable for items */
  SUPP   k;                     /* loop variable for points */
  size_t j, n;                  /* loop variables for points */
  double *p, t;                 /* to traverse the points */
  double span;                  /* length of the point span */
  double *pts, *s;              /* point population to sample from */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = clone(src))) return NULL; }
  if (dst != src)               /* if not identical to source, */
    tns_copy(dst, src);         /* copy the original train set */
  pts = s = tns_points(dst);    /* get an array with all points */
  if (!pts) return NULL;        /* to create a point permutation */
  for (n = tns_extent(src); --n > 0; ) {    /* shuffle the points */
    j = (size_t)(rng_dbl(rng) *(double)(n+1));
    if (j >= n) j = n;          /* get random index in {0,...,n} */
    t = s[j]; s[j] = *s; *s++ = t;
  }                             /* exchange the array elements */
  span = end -beg;              /* and compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  for (s = pts, i = 0; i < dst->cnt; i++) {
    p = dst->trains[i]->pts;    /* traverse the trains */
    dst->trains[i]->len = src->trains[i]->len;
    for (k = 0; k < dst->trains[i]->len; k++) {
      t = fmod(rng_next(rng) +(*s++ -beg), span);
      *p++ = t +((t < 0) ? end : beg);
    }                           /* sample from the kernel function */
    *p = TRN_END;               /* and wrap round if outside interval */
  }                             /* finally store a sentinel */
  dst->min =  INFINITY;         /* invalidate the range of points */
  dst->max = -INFINITY;         /* in the train set */
  return dst;                   /* return the created surrogate */
}  /* tns_permute() */

/*--------------------------------------------------------------------*/

TRAINSET* tns_permblk (TRAINSET *src, double beg, double end,
                       double delta, RNG *rng, TRAINSET *dst)
{                               /* --- blocked permutations */
  ITEM   i, n;                  /* loop variables for items */
  size_t j, k;                  /* loop variables for points */
  double **pos;                 /* position array for point output */
  double *pts, *s;              /* point population to permute */
  double span;                  /* length of the point span */
  double e, t;                  /* end of current block, swap buffer */

  assert(src && (beg <= end) && rng); /* check function arguments */
  if (!dst) { if (!(dst = clone(src))) return NULL; }
  if (dst != src)               /* if not identical to source, */
    tns_copy(dst, src);         /* copy the original train set */
  pts = s = tns_points(dst);    /* get an array with all points */
  if (!pts) return NULL;        /* to create a point permutation */
  n = tns_cnt(dst);             /* get the number of items */
  if (!dst->buf) {              /* if no position array created yet */
    dst->buf = malloc((size_t)n *sizeof(double*));
    if (!dst->buf) return NULL; /* create a position array */
  }                             /* for the permuted point output */
  pos = (double**)dst->buf;     /* get the position array */
  for (i = 0; i < n; i++)       /* collect the train starts */
    pos[i] = dst->trains[i]->pts;
  dbl_qsort(pts, tns_extent(dst), +1);     /* sort the points */
  span = end -(e = beg);        /* compute the span of points */
  if (span <= 0) span = 1.0;    /* ensure a proper span of points */
  while (*s < TRN_END) {        /* while more points to permute */
    e += delta;                 /* compute end of next block */
    while (*s < e) s++;         /* collect points in the block */
    if (s <= pts) continue;     /* check for an empty block */
    for (k = (size_t)(s-pts), s = pts; --k > 0; ){
      j = (size_t)(rng_dbl(rng) *(double)(k+1));
      if (j >= k) j = k;        /* get random index in {0,...,k} */
      t = s[j]; s[j] = *s; *s++ = t;
    }                           /* exchange the array elements */
    s++;                        /* (shuffle points in the block) */
    for (i = 0; i < n; i++) {   /* traverse the trains */
      while (*pos[i] < e) {     /* while inside the block */
        t = fmod(rng_next(rng) +(*pts++ -beg), span);
        *pos[i]++ = t +((t < 0) ? end : beg);
      }                         /* dither the shuffled points */
    }                           /* and store them in the trains */
    assert(pts == s);           /* check whether all points */
  }                             /* have been stored again */
  return dst;                   /* return the created surrogate */
}  /* tns_permblk() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/
#ifdef TRN_MAIN

static void help (void)
{                               /* --- print add. option information */
  #ifndef QUIET
  fprintf(stderr, "\n");        /* terminate startup message */
  printf("surrogate data generation methods (option -g#)\n");
  printf("  x    none (read pattern spectrum from file)\n");
  printf("  i    identity (keep original data)\n");
  printf("  r    point/spike time randomization\n");
  printf("  d    point/spike dithering/displacement\n");
  printf("  s    train shifting/dithering\n");
  printf("  k    sampling from a kernel estimate\n");
  printf("  p    dithered point/spike permutation\n");
  printf("  b    dithered blocked point/spike permutation\n");
  printf("random function for displacements (option -d#)\n");
  printf("  u/r  uniform/rectangular density\n");
  printf("  t    symmetric triangular density\n");
  printf("  g/n  Gaussian/normal density\n");
  printf("The sigma parameter states the standard deviation\n");
  printf("or half the base width of the rectangle/triangle.\n");
  #endif                        /* print help information */
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (rng)    rng_delete(rng);       \
  if (trnset) tns_delete(trnset, 0); \
  if (tread)  trd_delete(tread,  1); \
  if (ibase)  ib_delete (ibase);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables */
  char    *s;                   /* to traverse the options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *fn_sel  = NULL;      /* name of item selection file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *fmt     = "%g";      /* output format for points */
  int     surr     = 'x';       /* surrogate data generation method */
  int     rand     = 'u';       /* random function for displacements */
  double  sigma    =  0.005;    /* sigma parameter for random func. */
  double  delta    =  0.03;     /* sigma parameter for random func. */
  double  beg      =  INFINITY; /* beginning of point/time range */
  double  end      = -INFINITY; /* end       of point/time range */
  int     minp     = TRN_ITEMPT;/* mode for train set reading */
  int     mout     = TRN_ITEMPT;/* mode for train set writing */
  long    seed;                 /* seed for random numbers */
  ITEM    n;                    /* number of trains */
  size_t  x;                    /* number of spikes */
  clock_t t;                    /* timer for measurements */

  #ifndef QUIET                 /* if not quiet version */
  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-g#      surrogate data generation method         "
                    "(default: none)\n");
    printf("-d#      random function type for displacements   "
                    "(default: %c)\n", (char)rand);
    printf("-p#      sigma parameter for random function      "
                    "(default: %g)\n", sigma);
    printf("-B#      block size for blocked permutation       "
                    "(default: %g)\n", delta);
    printf("-S#      seed for random number generator         "
                    "(default: time)\n");
    printf("-a#      start of point/time range                "
                    "(default: auto)\n");
    printf("-z#      end   of point/time range                "
                    "(default: auto)\n");
    printf("-l       one train per record (item + points)     "
                    "(default: id/time pairs)\n");
    printf("-y       no item (with -l) or item after point    "
                    "(default: item first)\n");
    printf("-L       write one train per record (item + pts.) "
                    "(default: id/time pairs)\n");
    printf("-Y       write no item (-L) or item after point   "
                    "(default: item first)\n");
    printf("-s       sort output by points/time               "
                    "(default: by item)\n");
    printf("-R#      read an item selection from a file\n");
    printf("-o#      format for point output                  "
                    "(default: \"%s\")\n", fmt);
    printf("-r#      record/transaction separators            "
                    "(default: \"\\n\")\n");
    printf("-f#      field /item        separators            "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                       "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                       "
                    "(default: \"#\")\n");
    printf("-!       print additional option information\n");
    printf("infile   file to read trains from                 "
                    "[required]\n");
    printf("outfile  file to write (surrogate) trains to      "
                    "[optional]\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  #endif  /* #ifndef QUIET */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get a default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case '!': help();                    break;
          case 'g': surr   = *s++;             break;
          case 'd': rand   = *s++;             break;
          case 'p': sigma  = strtod(s, &s);    break;
          case 'B': delta  = strtod(s, &s);    break;
          case 'S': seed   = strtol(s, &s, 0); break;
          case 'a': beg    = strtod(s, &s);    break;
          case 'z': end    = strtod(s, &s);    break;
          case 'l': minp  &= ~TRN_ITEMPT;      break;
          case 'y': minp  |=  TRN_PURETRN;     break;
          case 'L': mout  &= ~TRN_ITEMPT;      break;
          case 'Y': mout  |=  TRN_PURETRN;     break;
          case 's': mout  |=  TRN_SORTED;      break;
          case 'R': optarg = &fn_sel;          break;
          case 'o': optarg = &fmt;             break;
          case 'r': optarg = &recseps;         break;
          case 'f': optarg = &fldseps;         break;
          case 'b': optarg = &blanks;          break;
          case 'C': optarg = &comment;         break;
          default : error(E_OPTION, *--s);     break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check (option) arguments */
  if (k < 1)  error(E_ARGCNT);  /* and number of arguments */
  if (strchr("urtgn",   rand) == NULL) error(E_RANDFN, (char)rand);
  if (strchr("xirdskp", surr) == NULL) error(E_SURRFN, (char)surr);
  MSG(stderr, "\n");            /* terminate the startup message */

  /* --- read item selection --- */
  ibase = ib_create(0, 0);      /* create an item base */
  if (!ibase) error(E_NOMEM);   /* to manage the items */
  tread = trd_create();         /* create a transaction reader */
  if (!tread) error(E_NOMEM);   /* and configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_sel) {                 /* if item appearances are given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_sel) != 0)
      error(E_FOPEN, trd_name(tread));
    MSG(stderr, "reading %s ... ", trd_name(tread));
    k = ib_readsel(ibase,tread);/* read the given item selection */
    if (k < 0) error(-k, ib_errmsg(ibase, NULL, 0));
    trd_close(tread);           /* close the input file */
    MSG(stderr, "[%"ITEM_FMT" item(s)]", ib_cnt(ibase));
    MSG(stderr, " done [%.2fs].", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- read train set --- */
  trnset = tns_create(ibase);   /* create a train set */
  if (!trnset) error(E_NOMEM);  /* to store the trains */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  MSG(stderr, "reading %s ... ", trd_name(tread));
  k = tns_read(trnset, tread, minp);
  if (k < 0)                    /* read the transaction database */
    error(-k, tns_errmsg(trnset, NULL, 0));
  trd_delete(tread, 1);         /* close the input file and */
  tread = NULL;                 /* delete the table reader */
  n = tns_cnt(trnset);          /* get the number of trains */
  x = tns_extent(trnset);       /* and the number of points */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" point(s)]", n, x);
  MSG(stderr, " done [%.2fs].", SEC_SINCE(t));
  if ((n <= 0) || (x <= 0))     /* check for at least one item */
    error(E_NOITEMS);           /* and at least one train */
  MSG(stderr, "\n");            /* terminate the log message */

  /* --- generate surrogate data --- */
  if ((beg <= -INFINITY) || (beg >= INFINITY))
    beg = floor(tns_min(trnset, 0));
  if ((beg <= -INFINITY) || (beg >= INFINITY))
    beg = 0.0;                  /* determine start from the data */
  if ((end <= -INFINITY) || (end >= INFINITY))
    end = ceil (tns_max(trnset, 0));
  if ((end <= -INFINITY) || (end >= INFINITY))
    end = 1.0;                  /* determine end   from the data */
  switch (rand) {               /* evaluate the random function id */
    case 'u': rand = RNG_UNIFORM;          break;
    case 'r': rand = RNG_RECT;             break;
    case 't': rand = RNG_TRIANG;           break;
    case 'g': rand = RNG_GAUSS;            break;
    case 'n': rand = RNG_NORMAL;           break;
    default : error(E_RANDFN, (char)rand); break;
  }                             /* get the symmetric random function */
  rng = rng_createx((unsigned int)seed, rand, sigma);
  if (!rng) error(E_NOMEM);     /* create a random number generator */
  switch (surr) {               /* evaluate the surrogate method id */
    case 'x': /* do nothing */                          break;
    case 'i': trnset = tns_ident  (trnset, beg, end,
                                   delta, rng, trnset); break;
    case 'r': trnset = tns_random (trnset, beg, end,
                                   delta, rng, trnset); break;
    case 'd': trnset = tns_dither (trnset, beg, end,
                                   delta, rng, trnset); break;
    case 's': trnset = tns_shift  (trnset, beg, end,
                                   delta, rng, trnset); break;
    case 'k': trnset = tns_kernest(trnset, beg, end,
                                   delta, rng, trnset); break;
    case 'p': trnset = tns_permute(trnset, beg, end,
                                   delta, rng, trnset); break;
    case 'b': trnset = tns_permblk(trnset, beg, end,
                                   delta, rng, trnset); break;
    default : error(E_SURRFN, (char)surr);              break;
  }                             /* generate a surrogate data set */
  rng_delete(rng); rng = NULL;  /* delete random number generator */

  /* --- sort points in trains --- */
  MSG(stderr, "sorting trains ... ");
  tns_sort(trnset, +1);         /* sort the trains by points */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" point(s)]", n, x);
  MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- write train set --- */
  if (fn_out) {                 /* if an output file is given */
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* configure the characters */
    twr_xchars(twrite, recseps, fldseps, blanks, "");
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    MSG(stderr, "writing %s ... ", twr_name(twrite));
    i = tns_write(trnset, twrite, mout, fmt);
    if (i < 0) error(i, twr_name(twrite));
    twr_delete(twrite, 0);      /* write the train data and */
    twrite = NULL;              /* delete the table writer */
    MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" point(s)]", n, x);
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif
